import sys
import os

cpu_index = 0
cpu_total = 1
job_index = 0

if len(sys.argv)>=3:
	cpu_index = int(sys.argv[1])
	cpu_total = int(sys.argv[2])
else:
	print("warning : arguments might be missing")

def add_job(func):
	global job_index
	if job_index == cpu_index:
		print(f"I'm running job {job_index}")
		func()
	job_index = (job_index+1)%cpu_total

###############################################################

from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import torch
import torch.nn as nn
import torch.nn.functional as F

import anyact as aa

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torchvision
import torchvision.datasets as datasets

import os
import h5py

from mnist import fmnist,mnist,mnist_perf,mnist_perf_light
import mnist_analysis

import matplotlib.pyplot as plt

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def expspace(x0,x1,N): return torch.exp(torch.linspace(np.log(x0),np.log(x1),N))

def gradiator(φ,grad_φ):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,x):
			y = φ(x)
			ctx.save_for_backward(x,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_o):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,y = ctx.saved_tensors # restore input from context
			grad_i = grad_φ(x,y,grad_o)
			return grad_i
	return GradiatorFunction.apply

###################################################################################
###################################################################################
###################################################################################

class DriftingLinear(nn.Module):
	def __init__(self,i,o,drift=0,noise=0):
		super().__init__()
		self.L = nn.Linear(i,o)
		self.drift_A_ = nn.parameter.Parameter(torch.randn(o,i),False)
		self.drift_b_ = nn.parameter.Parameter(torch.ones(o))
		self.drift_A = drift
		self.drift_b = drift
		self.noise_A = noise
		self.noise_b = noise
		# self.drifting = False
	def set_noise(self,noise):
		self.noise_A = noise
		self.noise_b = noise
		return self
	def set_drift(self,drift):
		self.drift_A = drift
		self.drift_b = drift
		return self
	def forward(self,x):
		# return self.L(x)
		drift_A = self.drift_A*self.drift_A_
		drift_b = self.drift_b*self.drift_b_
		noise_A = torch.randn(self.drift_A_.shape,device=self.drift_A_.device)*self.noise_A
		noise_b = torch.randn(self.drift_b_.shape,device=self.drift_b_.device)*self.noise_b
		A = self.L.weight * ( 1 + drift_A + noise_A)
		b = self.L.bias   * ( 1 + drift_b + noise_b)
		return torch.einsum("ij,Nj->Ni",A,x) +b

# device="cuda"
# L = DriftingLinear(10,20).to(device)
# x = torch.randn(3,10).to(device)
# L(x).shape

# L = nn.Sequential(
# 	DriftingLinear(10,20,noise=1,drift=0),
# 	DriftingLinear(20,10,noise=1,drift=0),
# )

# L = L.to(device)
# L(x)

class Sin(nn.Module):
	def __init__(self): super().__init__()
	def forward(self,x): return torch.sin(x)

class DriftClassifier(nn.Module):
	def __init__(self,layers):
		super().__init__()
		layers = [int(np.round(x)) for x in layers]
		self.L  = [DriftingLinear(λ1,λ2) for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.BN = [nn.BatchNorm1d(λ2)    for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		# self.σ  = [nn.GELU()             for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.σ  = [Sin()                 for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.model = nn.Sequential(*reduce(operator.add, zip(self.L,self.σ,self.BN))[:-2])
		self.softmax = nn.Softmax(dim=1)
	def set_noise(self,noise):
		for L in self.L: L.set_noise(noise)
		return self
	def set_drift(self,drift):
		for L in self.L: L.set_drift(drift)
		return self
	def forward(self,x):
		return self.softmax(self.model(nn.Flatten()(x)))

fmnist_L_x = mnist_analysis.fmnist_L_x
fmnist_L_y = mnist_analysis.fmnist_L_y
fmnist_T_x = mnist_analysis.fmnist_T_x
fmnist_T_y = mnist_analysis.fmnist_T_y

batch_size=64
rgb_L = (0,.5,1)
rgb_T = (1,.5,0)

# for N in [2,4,16]:
# N=4
# classifier = DriftClassifier(expspace(28*28,10,N+1))
# identifier = f"fc({N})"
# # mnist_analysis.analysis_classification(classifier, identifier, nn.MSELoss(),)
# mnist_analysis.analysis_classification(classifier, identifier, nn.MSELoss(),batch_size=2048)
#### bak = classifier

	# drift_min,drift_max=0,.25
	# noise_min,noise_max=0,.25
	# drift_N = 32
	# noise_N = 32
	# drifts = np.linspace(drift_min,drift_max,drift_N)
	# noises = np.linspace(noise_min,noise_max,noise_N)
	# accuracy  = np.zeros((N,drift_N,noise_N))
	# cm        = np.zeros((N,drift_N,noise_N,10,10))

	# batch_size=2**14
	# for drift_n,drift in enumerate(drifts):
	# 	print(drift_n)
	# 	for noise_n,noise in enumerate(noises):
	# 		classifier = classifier.set_noise(noise).set_drift(drift)
	# 		ac_T,cm_T=mnist_perf_light(classifier,device,f"drift={drift}-noise={noise}-test" ,fmnist_T_x,fmnist_T_y,batch_size,rgb_T) ### on test set
	# 		accuracy[n,drift_n,noise_n    ] = ac_T
	# 		cm      [n,drift_n,noise_n,:,:] = cm_T
	# plt.clf()
	# plt.imshow(accuracy, origin='lower', extent=(noise_min,noise_max,drift_min,drift_max))
	# plt.title("accuracy")
	# plt.xlabel("noise")
	# plt.ylabel("drift")
	# plt.colorbar()





noise_min,noise_max=0,.25
noise_N = 512
noises = np.linspace(noise_min,noise_max,noise_N)
Ns = [2,4,8,16]
accuracy2  = np.zeros((len(Ns),noise_N,))
cm2        = np.zeros((len(Ns),noise_N,10,10))
for n,N in enumerate(Ns):
	classifier = DriftClassifier(expspace(28*28,10,N+1))
	identifier = f"fc({N})"
	mnist_analysis.analysis_classification(classifier, identifier, nn.MSELoss(),batch_size=2048,lr0=-2,lr1=-6)
	batch_size=2**14
	for noise_n,noise in enumerate(noises):
		classifier = classifier.set_noise(noise).set_drift(0)
		ac_T,cm_T=mnist_perf_light(classifier,device,f"noise={noise:.6f}-test" ,fmnist_T_x,fmnist_T_y,batch_size,rgb_T) ### on test set
		accuracy2[n,noise_n    ] = ac_T
		cm2      [n,noise_n,:,:] = cm_T
	with h5py.File(f'out/{identifier}-drifting.h5', 'w') as file:
		file.create_dataset("accuracy", data=accuracy2[n])
		file.create_dataset("cm", data=cm2[n])
		file.create_dataset("noises", data=noises)

plt.close('all')
for n,N in enumerate(Ns):
	plt.scatter(noises,accuracy2[n],s=1,label=f"fc({N})")

# plt.tight_layout()
plt.legend()
plt.xlabel("$σ$")
plt.ylabel("accuracy")
plt.savefig(f"out/drifting-all-multiplicative_noise-sin.png")
# $θ_i + \mathcal{N}(σ,0)$

plt.close('all')
for n,N in enumerate(Ns):
	plt.scatter(noises,np.log(1-accuracy2[n]),s=1,label=f"fc({N})")

# plt.tight_layout()
plt.legend()
plt.xlabel("$σ$")
plt.ylabel("log(1-accuracy)")
plt.savefig(f"out/drifting-all-multiplicative_noise-sin-v2.png")















noise_min,noise_max=0,.25
noise_N = 512
noises = np.linspace(noise_min,noise_max,noise_N)
Ns = [2,4,8,16]
accuracy2  = np.zeros((len(Ns),noise_N,))
cm2        = np.zeros((len(Ns),noise_N,10,10))
for n,N in enumerate(Ns):
	identifier = f"fc({N})"
	with h5py.File(f'out/{identifier}-drifting-{N}.h5', 'r') as file:
		# A = np.array(file["accuracy"])[n]
		accuracy2[n] = np.array(file["accuracy"])[n+1]
		cm2      [n] = np.array(file["cm"])[n+1]

plt.clf()
for n,N in enumerate(Ns):
	plt.scatter(noises,accuracy2[n],s=1,label=f"fc({N})")
plt.tight_layout()
plt.legend()
plt.savefig(f"out/{identifier}-drifting-all.png")


